/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */

package hu.mta.sztaki.lpds.cloud.simulator.storage;

import hu.mta.sztaki.lpds.cloud.simulator.DeferredEvent;
import hu.mta.sztaki.lpds.cloud.simulator.Timed;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.PriorityQueue;

public class NetworkNode {

	public static class NetworkException extends Exception {
		private static final long serialVersionUID = 5173643896341066497L;

		public NetworkException(String msg) {
			super(msg);
		}
	}

	public static interface NetworkEvent {
		public void trComplete();
	}

	static class SingleTransfer extends DeferredEvent {

		public long currentContent;
		public long totalTransfer;
		final public NetworkEvent ev;

		public final BandwidthSpreader in;
		public final BandwidthSpreader out;

		private SingleTransfer(final int latency, final long tottr,
				final BandwidthSpreader in, final BandwidthSpreader out,
				final NetworkEvent e) {
			super(latency);
			currentContent = 0;
			totalTransfer = tottr;
			ev = e;
			this.in = in;
			this.out = out;
		}

		@Override
		protected void eventAction() {
			out.registerTR(this);
			in.registerTR(this);
		}

		public long getRemainingTransfer() {
			return currentContent + totalTransfer;
		}

		@Override
		public String toString() {
			return "ST(C:" + currentContent + " T:" + totalTransfer + ")";
		}

	}

	static final Comparator<SingleTransfer> stComp = new Comparator<SingleTransfer>() {
		@Override
		public int compare(final SingleTransfer o1, final SingleTransfer o2) {
			final long trOth = o1.getRemainingTransfer();
			final long trThis = o2.getRemainingTransfer();
			return trOth < trThis ? -1 : (trOth == trThis ? 0 : 1);
		}

		@Override
		public boolean equals(Object obj) {
			System.err.println("WHOO!");
			System.exit(0);
			return false;
		}
	};

	private abstract class BandwidthSpreader extends Timed {
		protected final long baseBW;
		protected final PriorityQueue<SingleTransfer> tg = new PriorityQueue<SingleTransfer>(
				500, stComp);

		protected long lastNotifTime = 0;
		protected long timeAdjustedBaseBW;

		private class SpreadingRecord {
			public long timestamp;
			public long totalTransferred;
		}

		private LinkedList<SpreadingRecord> subSecondRecords = new LinkedList<SpreadingRecord>();
		private LinkedList<SpreadingRecord> subHourRecords = new LinkedList<SpreadingRecord>();
		private LinkedList<SpreadingRecord> subDayRecords = new LinkedList<SpreadingRecord>();
		private long subSecondTransfers = 0;
		private long subHourTransfers = 0;
		private long subDayTransfers = 0;
		private long totalTransferred = 0;

		public BandwidthSpreader(final long initialBW) {
			baseBW = initialBW;
		}

		protected long calcNewFreq() {
			final int tgs = tg.size();
			if (tgs > 0) {
				final long minimum = tg.peek().getRemainingTransfer();
				return 1 + (long) (minimum * 1000 / ((double) baseBW / tgs));
			}
			return 1;
		}

		protected void updateForAllTrs(long newfreq) {
			frequpdate: while (true) {
				final long oldNe = getNextEvent();
				final long neOur = updateFrequency(newfreq);
				if (oldNe == neOur) {
					return;
				}
				for (final SingleTransfer st : tg) {
					final BandwidthSpreader counterPart = getCounterPart(st);
					final long neCP = counterPart.getNextEvent();
					if (neOur < neCP) {
						counterPart.updateForAllTrs(newfreq);
					} else if (neOur != neCP && counterPart.isSubscribed()) {
						final long nfupd = counterPart.nextEventDistance();
						if (nfupd != 0) {
							newfreq = nfupd;
							continue frequpdate;
						}
					}
				}
				return;
			}
		}

		public void registerTR(final SingleTransfer st) {
			tg.offer(st);
			final long nf = calcNewFreq();
			if (isSubscribed()) {
				updateForAllTrs(nf);
				return;
			} else {
				lastNotifTime = Timed.getFireCount();
				subscribe(nf);
			}
		}

		@Override
		public void tick(final long currentFireCount) {
			final double elapsedFires = currentFireCount - lastNotifTime;
			long remainingBandwidth = timeAdjustedBaseBW = 1 + (long) (baseBW * (elapsedFires / 1000));
			long pastRemaining;
			try {
				do {
					pastRemaining = remainingBandwidth;
					remainingBandwidth = singleSpread(pastRemaining / tg.size());

				} while (remainingBandwidth > 0
						&& pastRemaining != remainingBandwidth);
			} catch (ArithmeticException e) {
			}
			totalTransferred += timeAdjustedBaseBW - remainingBandwidth;
			updateRecords(currentFireCount);

			if (tg.size() == 0) {
				unsubscribe();
				return;
			} else {
				updateForAllTrs(calcNewFreq());
				lastNotifTime = currentFireCount;
			}
		}

		private void updateRecords(final long timeStamp) {
			final SpreadingRecord current = new SpreadingRecord();
			current.timestamp = timeStamp;
			current.totalTransferred = totalTransferred;
			subSecondRecords.add(current);
			while (subSecondRecords.peek().timestamp + 1000 < timeStamp) {
				subHourRecords.add(subSecondRecords.poll());
				while (subHourRecords.peek().timestamp + 3600000 < timeStamp) {
					subDayRecords.add(subHourRecords.poll());
					while (subDayRecords.peek().timestamp + 86400000 < timeStamp) {
						subDayRecords.poll();
					}
				}
			}
			subSecondTransfers = totalTransferred
					- subSecondRecords.peek().totalTransferred;
			try {
				subHourTransfers = totalTransferred
						- subHourRecords.peek().totalTransferred;
			} catch (NullPointerException ne) {
				subHourTransfers = subSecondTransfers;
			}
			try {
				subDayTransfers = totalTransferred
						- subDayRecords.peek().totalTransferred;
			} catch (NullPointerException ne) {
				subDayTransfers = subHourTransfers;
			}
		}

		public long getTotalTransferred() {
			return totalTransferred;
		}

		public long getSubDayTransfers() {
			return subDayTransfers;
		}

		public long getSubHourTransfers() {
			return subHourTransfers;
		}

		public long getSubSecondTransfers() {
			return subSecondTransfers;
		}

		protected abstract long singleSpread(final long toSpread);

		protected abstract BandwidthSpreader getCounterPart(
				final SingleTransfer st);
	}

	private class InputSpreader extends BandwidthSpreader {

		public InputSpreader(final long initialBW) {
			super(initialBW);
		}

		@Override
		public int compareTo(Timed o) {
			final int returner = super.compareTo(o);
			return ((returner == 0) && (o instanceof OutputSpreader)) ? 1
					: returner;
		}

		@Override
		protected BandwidthSpreader getCounterPart(final SingleTransfer st) {
			return st.out;
		}

		@Override
		protected long singleSpread(final long toSpread) {
			long remainingBandwidth = 0;
			final Iterator<SingleTransfer> i = tg.iterator();
			while (i.hasNext()) {
				final SingleTransfer st = i.next();
				final long reducedContent = st.currentContent - toSpread;
				if (reducedContent > 0) {
					st.currentContent = reducedContent;
					continue;
				} else {
					remainingBandwidth -= reducedContent;
					if (st.totalTransfer == 0) {
						st.ev.trComplete();
						i.remove();
						continue;
					} else {
						st.currentContent = 0;
					}
				}
			}
			return remainingBandwidth;
		}

		@Override
		public String toString() {
			return "IS(" + name + "," + tg + ")";
		}
	}

	private class OutputSpreader extends BandwidthSpreader {
		public OutputSpreader(final long initialBW) {
			super(initialBW);
		}

		@Override
		public int compareTo(Timed o) {
			final int returner = super.compareTo(o);
			return ((returner == 0) && (o instanceof InputSpreader)) ? -1
					: returner;
		}

		@Override
		protected BandwidthSpreader getCounterPart(final SingleTransfer st) {
			return st.in;
		}

		@Override
		protected long singleSpread(final long toSpread) {
			long remainingBandwidth = toSpread * tg.size();
			final Iterator<SingleTransfer> i = tg.iterator();
			while (i.hasNext()) {
				final SingleTransfer st = i.next();
				final long realRequestedContent = st.currentContent
						+ (toSpread < st.totalTransfer ? toSpread
								: st.totalTransfer);
				final long cc = realRequestedContent < timeAdjustedBaseBW ? realRequestedContent
						: timeAdjustedBaseBW;
				final long transferdiff = cc - st.currentContent;
				st.currentContent = cc;
				st.totalTransfer -= transferdiff;

				if (st.totalTransfer == 0) {
					i.remove();
				}

				remainingBandwidth -= transferdiff;
			}
			return remainingBandwidth;
		}

		@Override
		public String toString() {
			return "OS(" + name + "," + tg + ")";
		}
	}

	private final BandwidthSpreader inbws;
	private final BandwidthSpreader outbws;
	private final BandwidthSpreader diskinbws;
	private final BandwidthSpreader diskoutbws;

	private final String name;
	private final HashMap<String, Integer> latencies;

	public NetworkNode(final String id, final long maxInBW,
			final long maxOutBW, final long diskBW,
			final HashMap<String, Integer> latencymap) {
		name = id;
		outbws = new OutputSpreader(maxOutBW);
		inbws = new InputSpreader(maxInBW);
		diskinbws = new InputSpreader(diskBW / 2);
		diskoutbws = new OutputSpreader(diskBW / 2);
		latencies = latencymap;
	}

	public long getOutputbw() {
		return outbws.baseBW;
	}

	public long getInputbw() {
		return inbws.baseBW;
	}

	public long getDiskbw() {
		return diskinbws.baseBW * 2;
	}

	public static void initTransfer(final long size, final NetworkNode from,
			final NetworkNode to, final NetworkEvent e) throws NetworkException {
		if (from == to) {
			new SingleTransfer(checkConnectivity(from, to), size,
					from.diskinbws, from.diskoutbws, e);
			return;
		} else {
			new SingleTransfer(checkConnectivity(from, to), size, to.inbws,
					from.outbws, e);
		}
	}

	public static int checkConnectivity(final NetworkNode from,
			final NetworkNode to) throws NetworkException {
		final Integer lat = from.latencies.get(to.name);
		if (lat == null)
			throw new NetworkException("No connection between: '" + from.name
					+ "' and '" + to.name + "'");
		return lat;
	}

	public String getName() {
		return name;
	}

	@Override
	public String toString() {
		return "NetworkNode: " + name + "(NI:" + getInputbw() + ",NO:"
				+ getOutputbw() + " -- RX:" + inbws.getTotalTransferred()
				+ " TX:" + outbws.getTotalTransferred() + "--, D:"
				+ getDiskbw() + ")";
	}

}
